@@ -27,6 +27,7 @@
 #include "util.h"
 #include "core.h"
 #include "themewidget.h"
+#include "metaaccellabel.h"
 
 typedef struct _MenuItem MenuItem;
 typedef struct _MenuData MenuData;
@@ -48,15 +49,16 @@ struct _MenuData
 static void activate_cb (GtkWidget *menuitem, gpointer data);
 
 static MenuItem menuitems[] = {
-  { META_MENU_OP_DELETE, NULL, N_("_Close") },
-  { META_MENU_OP_MINIMIZE, NULL, N_("_Minimize") },
-  { META_MENU_OP_MAXIMIZE, NULL, N_("Ma_ximize") },
+  { META_MENU_OP_MINIMIZE, METACITY_STOCK_MINIMIZE, N_("Mi_nimize") },
+  { META_MENU_OP_MAXIMIZE, METACITY_STOCK_MAXIMIZE, N_("Ma_ximize") },
   { META_MENU_OP_UNMAXIMIZE, NULL, N_("Unma_ximize") },
-  { META_MENU_OP_SHADE, NULL, N_("_Shade") },
-  { META_MENU_OP_UNSHADE, NULL, N_("Un_shade") },
-  { META_MENU_OP_MOVE, NULL, N_("Mo_ve") },
+  { META_MENU_OP_SHADE, NULL, N_("Roll _Up") },
+  { META_MENU_OP_UNSHADE, NULL, N_("_Unroll") },
+  { META_MENU_OP_MOVE, NULL, N_("_Move") },
   { META_MENU_OP_RESIZE, NULL, N_("_Resize") },
   { 0, NULL, NULL }, /* separator */
+  { META_MENU_OP_DELETE, METACITY_STOCK_DELETE, N_("_Close") },
+  { 0, NULL, NULL }, /* separator */
   { META_MENU_OP_STICK, NULL, N_("Put on _All Workspaces") },
   { META_MENU_OP_UNSTICK, NULL, N_("Only on _This Workspace") }
 };
@@ -120,40 +122,6 @@ activate_cb (GtkWidget *menuitem, gpointer data)
   /* menu may now be freed */
 }
 
-static void
-menu_icon_size_func (MetaArea *area,
-                     int      *width,
-                     int      *height,
-                     void     *user_data)
-{
-  gtk_icon_size_lookup (GTK_ICON_SIZE_MENU,
-                        width, height);
-}
-
-static void
-menu_icon_expose_func (MetaArea       *area,
-                       GdkEventExpose *event,
-                       int             x_offset,
-                       int             y_offset,
-                       void           *user_data)
-{
-  int width, height;
-  MetaMenuIconType type;
-  
-  type = GPOINTER_TO_INT (user_data);
-  
-  gtk_icon_size_lookup (GTK_ICON_SIZE_MENU,
-                        &width, &height);
-  
-  meta_theme_draw_menu_icon (meta_theme_get_current (),
-                             GTK_WIDGET (area),
-                             GTK_WIDGET (area)->window,
-                             &event->area,
-                             x_offset, y_offset,
-                             width, height,
-                             type);
-}
-
 /*
  * Given a Display and an index, get the workspace name and add any
  * accelerators. At the moment this means adding a _ if the name is of
@@ -163,30 +131,36 @@ menu_icon_expose_func (MetaArea       *area,
  * The calling code owns the string, and is reponsible to free the
  * memory after use.
  */
-static char *
+static char*
 get_workspace_name_with_accel (Display *display,
-                               int index)
+                               Window   xroot,
+                               int      index)
 {
-  char *name;
-  unsigned int number;
+  const char *name;
+  int number;
 
-  name = meta_core_get_workspace_name_with_index (display, index);
+  name = meta_core_get_workspace_name_with_index (display, xroot, index);
 
+  g_assert (name != NULL);
+  
   /*
    * If the name is of the form "Workspace x" where x is an unsigned
    * integer, insert a '_' before the number if it is less than 10 and
    * return it
    */
-  if (sscanf (name, _("Workspace %u"), &number) == 1)
+  number = 0;
+  if (sscanf (name, _("Workspace %d"), &number) == 1)
     {
+      char *new_name;
+      
       /*
        * Above name is a pointer into the Workspace struct. Here we make
        * a copy copy so we can have our wicked way with it.
        */
-      name = g_strdup_printf (_("Workspace %s%d"),
-                              number < 10 ? "_" : "",
-                              number);
-      return name;
+      new_name = g_strdup_printf (_("Workspace %s%d"),
+                                  number < 10 ? "_" : "",
+                                  number);
+      return new_name;
     }
   else
     {
@@ -195,27 +169,54 @@ get_workspace_name_with_accel (Display *display,
        * add accelerators. Escape any _ characters so that the user's
        * workspace names do not get mangled.
        */
-      char *new_name, *source, *dest;
-      source = name;
+      char *new_name;
+      const char *source;
+      char *dest;
+
       /*
        * Assume the worst case, that every character is a _
        */
-      dest = new_name = g_malloc0 (strlen (name) * 2);
+      new_name = g_malloc0 (strlen (name) * 2 + 1);
+
       /*
        * Now iterate down the strings, adding '_' to escape as we go
        */
+      dest = new_name;
+      source = name;
       while (*source != '\0')
         {
           if (*source == '_')
             *dest++ = '_';
           *dest++ = *source++;
         }
-      /*
-       * We don't free *name as we don't own it, and pass ownership of
-       * *new_name to the calling code.
-       */
+
       return new_name;
-  }
+    }
+}
+
+static GtkWidget*
+menu_item_new (const char         *label,
+               gboolean            with_image,
+               unsigned int        key,
+               MetaVirtualModifier mods)
+{
+  GtkWidget *menu_item;
+  GtkWidget *accel_label;
+
+  if (with_image)
+    menu_item = gtk_image_menu_item_new ();
+  else
+    menu_item = gtk_menu_item_new ();
+  accel_label = meta_accel_label_new_with_mnemonic (label);
+  gtk_misc_set_alignment (GTK_MISC (accel_label), 0.0, 0.5);
+
+  gtk_container_add (GTK_CONTAINER (menu_item), accel_label);
+  gtk_widget_show (accel_label);
+
+  meta_accel_label_set_accelerator (META_ACCEL_LABEL (accel_label),
+                                    key, mods);
+  
+  return menu_item;
 }
 
 MetaWindowMenu*
@@ -240,10 +241,10 @@ meta_window_menu_new   (MetaFrames         *frames,
   menu->insensitive = insensitive;
   
   menu->menu = gtk_menu_new ();
-#ifdef HAVE_GTK_MULTIHEAD
+
   gtk_menu_set_screen (GTK_MENU (menu->menu),
 		       gtk_widget_get_screen (GTK_WIDGET (frames)));
-#endif
+
   i = 0;
   while (i < (int) G_N_ELEMENTS (menuitems))
     {
@@ -251,7 +252,9 @@ meta_window_menu_new   (MetaFrames         *frames,
         {
           GtkWidget *mi;
           MenuData *md;
-
+          unsigned int key;
+          MetaVirtualModifier mods;
+          
           if (menuitems[i].op == 0)
             {
               mi = gtk_separator_menu_item_new ();
@@ -262,65 +265,26 @@ meta_window_menu_new   (MetaFrames         *frames,
 
               image = NULL;
               
-              switch (menuitems[i].op)
-                {
-                case META_MENU_OP_MAXIMIZE:
-                  image = meta_area_new ();
-                  meta_area_setup (META_AREA (image),
-                                   menu_icon_size_func,
-                                   menu_icon_expose_func,
-                                   GINT_TO_POINTER (META_MENU_ICON_TYPE_MAXIMIZE),
-                                   NULL);
-                  break;
-
-                case META_MENU_OP_UNMAXIMIZE:
-                  image = meta_area_new ();
-                  meta_area_setup (META_AREA (image),
-                                   menu_icon_size_func,
-                                   menu_icon_expose_func,
-                                   GINT_TO_POINTER (META_MENU_ICON_TYPE_UNMAXIMIZE),
-                                   NULL);
-                  break;
-                  
-                case META_MENU_OP_MINIMIZE:
-                  image = meta_area_new ();
-                  meta_area_setup (META_AREA (image),
-                                   menu_icon_size_func,
-                                   menu_icon_expose_func,
-                                   GINT_TO_POINTER (META_MENU_ICON_TYPE_MINIMIZE),
-                                   NULL);
-                  break;
-
-                case META_MENU_OP_DELETE:
-                  image = meta_area_new ();
-                  meta_area_setup (META_AREA (image),
-                                   menu_icon_size_func,
-                                   menu_icon_expose_func,
-                                   GINT_TO_POINTER (META_MENU_ICON_TYPE_CLOSE),
-                                   NULL);
-                  break;
-                default:
-                  break;
-                }
-              
-              if (image == NULL && 
-                  menuitems[i].stock_id)
+              if (menuitems[i].stock_id)
                 {
                   image = gtk_image_new_from_stock (menuitems[i].stock_id,
                                                     GTK_ICON_SIZE_MENU);
 
                 }
+
+              meta_core_get_menu_accelerator (menuitems[i].op, -1,
+                                              &key, &mods);
               
               if (image)
                 {
-                  mi = gtk_image_menu_item_new_with_mnemonic (_(menuitems[i].label));
+                  mi = menu_item_new (_(menuitems[i].label), TRUE, key, mods);
                   gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (mi),
                                                  image);
                   gtk_widget_show (image);
                 }
               else
                 {
-                  mi = gtk_menu_item_new_with_mnemonic (_(menuitems[i].label));
+                  mi = menu_item_new (_(menuitems[i].label), FALSE, key, mods);
                 }
               
               if (insensitive & menuitems[i].op)
@@ -352,25 +316,36 @@ meta_window_menu_new   (MetaFrames         *frames,
       meta_verbose ("Creating %d-workspace menu current space %d\n",
                     n_workspaces, active_workspace);
       
-      if (n_workspaces > 0)
+      if (n_workspaces > 1)
         {
           GtkWidget *mi;
           Display *display;
-
-          display = gdk_x11_drawable_get_xdisplay(GTK_WIDGET(frames)->window);
-
+          Window xroot;
+          GdkScreen *screen;
+            
+          display = gdk_x11_drawable_get_xdisplay (GTK_WIDGET (frames)->window);
+          
+          screen = gdk_drawable_get_screen (GTK_WIDGET (frames)->window);
+          xroot = GDK_DRAWABLE_XID (gdk_screen_get_root_window (screen));
+          
           i = 0;
           while (i < n_workspaces)
             {
               char *label, *name;
               MenuData *md;
-
-              name = get_workspace_name_with_accel (display, i);
+              unsigned int key;
+              MetaVirtualModifier mods;
+              
+              meta_core_get_menu_accelerator (META_MENU_OP_WORKSPACES,
+                                              i + 1,
+                                              &key, &mods);
+              
+              name = get_workspace_name_with_accel (display, xroot, i);
               if (ops & META_MENU_OP_UNSTICK)
                 label = g_strdup_printf (_("Only on %s"), name);
               else
                 label = g_strdup_printf(_("Move to %s"), name);
-              mi = gtk_menu_item_new_with_mnemonic (label);
+              mi = menu_item_new (label, FALSE, key, mods);
 
               g_free (name);
               g_free (label);
